@@ -5,45 +5,23 @@
 //
 // This file may be distributed under the terms of the GNU LGPLv3 license.
 
-#include "disk.h" // floppy_13
 #include "biosvar.h" // SET_BDA
-#include "config.h" // CONFIG_*
-#include "util.h" // debug_enter
-#include "pic.h" // eoi_pic2
 #include "bregs.h" // struct bregs
-#include "pci.h" // pci_bdf_to_bus
-#include "ata.h" // ATA_CB_DC
+#include "config.h" // CONFIG_*
+#include "hw/ata.h" // ATA_CB_DC
+#include "hw/pci.h" // pci_bdf_to_bus
+#include "hw/pic.h" // pic_eoi2
+#include "output.h" // debug_enter
+#include "stacks.h" // call16_int
+#include "std/disk.h" // DISK_RET_SUCCESS
+#include "string.h" // memset
+#include "util.h" // CDRom_locks
 
 
 /****************************************************************
  * Helper functions
  ****************************************************************/
 
-void
-__disk_ret(struct bregs *regs, u32 linecode, const char *fname)
-{
-    u8 code = linecode;
-    if (regs->dl < EXTSTART_HD)
-        SET_BDA(floppy_last_status, code);
-    else
-        SET_BDA(disk_last_status, code);
-    if (code)
-        __set_code_invalid(regs, linecode, fname);
-    else
-        set_code_success(regs);
-}
-
-void
-__disk_ret_unimplemented(struct bregs *regs, u32 linecode, const char *fname)
-{
-    u8 code = linecode;
-    if (regs->dl < EXTSTART_HD)
-        SET_BDA(floppy_last_status, code);
-    else
-        SET_BDA(disk_last_status, code);
-    __set_code_unimplemented(regs, linecode, fname);
-}
-
 static void
 __disk_stub(struct bregs *regs, int lineno, const char *fname)
 {
@@ -55,32 +33,32 @@ __disk_stub(struct bregs *regs, int lineno, const char *fname)
     __disk_stub((regs), __LINE__, __func__)
 
 // Get the cylinders/heads/sectors for the given drive.
-static void
-fillLCHS(struct drive_s *drive_g, u16 *nlc, u16 *nlh, u16 *nlspt)
+static struct chs_s
+getLCHS(struct drive_s *drive_gf)
 {
-    if (CONFIG_CDROM_EMU
-        && drive_g == GLOBALFLAT2GLOBAL(GET_GLOBAL(cdemu_drive_gf))) {
-        // Emulated drive - get info from ebda.  (It's not possible to
+    struct chs_s res = { };
+    if (CONFIG_CDROM_EMU && drive_gf == GET_GLOBAL(cdemu_drive_gf)) {
+        // Emulated drive - get info from CDEmu.  (It's not possible to
         // populate the geometry directly in the driveid because the
         // geometry is only known after the bios segment is made
         // read-only).
-        u16 ebda_seg = get_ebda_seg();
-        *nlc = GET_EBDA2(ebda_seg, cdemu.lchs.cylinders);
-        *nlh = GET_EBDA2(ebda_seg, cdemu.lchs.heads);
-        *nlspt = GET_EBDA2(ebda_seg, cdemu.lchs.spt);
-        return;
+        res.cylinder = GET_LOW(CDEmu.lchs.cylinder);
+        res.head = GET_LOW(CDEmu.lchs.head);
+        res.sector = GET_LOW(CDEmu.lchs.sector);
+        return res;
     }
-    *nlc = GET_GLOBAL(drive_g->lchs.cylinders);
-    *nlh = GET_GLOBAL(drive_g->lchs.heads);
-    *nlspt = GET_GLOBAL(drive_g->lchs.spt);
+    res.cylinder = GET_GLOBALFLAT(drive_gf->lchs.cylinder);
+    res.head = GET_GLOBALFLAT(drive_gf->lchs.head);
+    res.sector = GET_GLOBALFLAT(drive_gf->lchs.sector);
+    return res;
 }
 
 // Perform read/write/verify using old-style chs accesses
 static void noinline
-basic_access(struct bregs *regs, struct drive_s *drive_g, u16 command)
+basic_access(struct bregs *regs, struct drive_s *drive_gf, u16 command)
 {
     struct disk_op_s dop;
-    dop.drive_g = drive_g;
+    dop.drive_gf = drive_gf;
     dop.command = command;
 
     u8 count = regs->al;
@@ -95,18 +73,18 @@ basic_access(struct bregs *regs, struct drive_s *drive_g, u16 command)
     }
     dop.count = count;
 
-    u16 nlc, nlh, nlspt;
-    fillLCHS(drive_g, &nlc, &nlh, &nlspt);
+    struct chs_s chs = getLCHS(drive_gf);
+    u16 nlc=chs.cylinder, nlh=chs.head, nls=chs.sector;
 
     // sanity check on cyl heads, sec
-    if (cylinder >= nlc || head >= nlh || sector > nlspt) {
+    if (cylinder >= nlc || head >= nlh || sector > nls) {
         warn_invalid(regs);
         disk_ret(regs, DISK_RET_EPARAM);
         return;
     }
 
     // translate lchs to lba
-    dop.lba = (((((u32)cylinder * (u32)nlh) + (u32)head) * (u32)nlspt)
+    dop.lba = (((((u32)cylinder * (u32)nlh) + (u32)head) * (u32)nls)
                + (u32)sector - 1);
 
     dop.buf_fl = MAKE_FLATPTR(regs->es, regs->bx);
@@ -120,21 +98,22 @@ basic_access(struct bregs *regs, struct drive_s *drive_g, u16 command)
 
 // Perform read/write/verify using new-style "int13ext" accesses.
 static void noinline
-extended_access(struct bregs *regs, struct drive_s *drive_g, u16 command)
+extended_access(struct bregs *regs, struct drive_s *drive_gf, u16 command)
 {
     struct disk_op_s dop;
+    struct int13ext_s *param_far = (struct int13ext_s*)(regs->si+0);
     // Get lba and check.
-    dop.lba = GET_INT13EXT(regs, lba);
+    dop.lba = GET_FARVAR(regs->ds, param_far->lba);
     dop.command = command;
-    dop.drive_g = drive_g;
-    if (dop.lba >= GET_GLOBAL(drive_g->sectors)) {
+    dop.drive_gf = drive_gf;
+    if (dop.lba >= GET_GLOBALFLAT(drive_gf->sectors)) {
         warn_invalid(regs);
         disk_ret(regs, DISK_RET_EPARAM);
         return;
     }
 
-    dop.buf_fl = SEGOFF_TO_FLATPTR(GET_INT13EXT(regs, data));
-    dop.count = GET_INT13EXT(regs, count);
+    dop.buf_fl = SEGOFF_TO_FLATPTR(GET_FARVAR(regs->ds, param_far->data));
+    dop.count = GET_FARVAR(regs->ds, param_far->count);
     if (! dop.count) {
         // Nothing to do.
         disk_ret(regs, DISK_RET_SUCCESS);
@@ -143,7 +122,7 @@ extended_access(struct bregs *regs, struct drive_s *drive_g, u16 command)
 
     int status = send_disk_op(&dop);
 
-    SET_INT13EXT(regs, count, dop.count);
+    SET_FARVAR(regs->ds, param_far->count, dop.count);
 
     disk_ret(regs, status);
 }
@@ -155,10 +134,10 @@ extended_access(struct bregs *regs, struct drive_s *drive_g, u16 command)
 
 // disk controller reset
 static void
-disk_1300(struct bregs *regs, struct drive_s *drive_g)
+disk_1300(struct bregs *regs, struct drive_s *drive_gf)
 {
     struct disk_op_s dop;
-    dop.drive_g = drive_g;
+    dop.drive_gf = drive_gf;
     dop.command = CMD_RESET;
     int status = send_disk_op(&dop);
     disk_ret(regs, status);
@@ -166,7 +145,7 @@ disk_1300(struct bregs *regs, struct drive_s *drive_g)
 
 // read disk status
 static void
-disk_1301(struct bregs *regs, struct drive_s *drive_g)
+disk_1301(struct bregs *regs, struct drive_s *drive_gf)
 {
     u8 v;
     if (regs->dl < EXTSTART_HD)
@@ -181,47 +160,48 @@ disk_1301(struct bregs *regs, struct drive_s *drive_g)
 
 // read disk sectors
 static void
-disk_1302(struct bregs *regs, struct drive_s *drive_g)
+disk_1302(struct bregs *regs, struct drive_s *drive_gf)
 {
-    basic_access(regs, drive_g, CMD_READ);
+    basic_access(regs, drive_gf, CMD_READ);
 }
 
 // write disk sectors
 static void
-disk_1303(struct bregs *regs, struct drive_s *drive_g)
+disk_1303(struct bregs *regs, struct drive_s *drive_gf)
 {
-    basic_access(regs, drive_g, CMD_WRITE);
+    basic_access(regs, drive_gf, CMD_WRITE);
 }
 
 // verify disk sectors
 static void
-disk_1304(struct bregs *regs, struct drive_s *drive_g)
+disk_1304(struct bregs *regs, struct drive_s *drive_gf)
 {
-    basic_access(regs, drive_g, CMD_VERIFY);
+    basic_access(regs, drive_gf, CMD_VERIFY);
 }
 
 // format disk track
 static void noinline
-disk_1305(struct bregs *regs, struct drive_s *drive_g)
+disk_1305(struct bregs *regs, struct drive_s *drive_gf)
 {
     debug_stub(regs);
 
-    u16 nlc, nlh, nlspt;
-    fillLCHS(drive_g, &nlc, &nlh, &nlspt);
+    struct chs_s chs = getLCHS(drive_gf);
+    u16 nlc=chs.cylinder, nlh=chs.head, nls=chs.sector;
 
-    u8 num_sectors = regs->al;
-    u8 head        = regs->dh;
+    u8 count = regs->al;
+    u8 cylinder = regs->ch;
+    u8 head = regs->dh;
 
-    if (head >= nlh || num_sectors == 0 || num_sectors > nlspt) {
+    if (cylinder >= nlc || head >= nlh || count == 0 || count > nls) {
         disk_ret(regs, DISK_RET_EPARAM);
         return;
     }
 
     struct disk_op_s dop;
-    dop.drive_g = drive_g;
+    dop.drive_gf = drive_gf;
     dop.command = CMD_FORMAT;
-    dop.lba = head;
-    dop.count = num_sectors;
+    dop.lba = (((u32)cylinder * (u32)nlh) + (u32)head) * (u32)nls;
+    dop.count = count;
     dop.buf_fl = MAKE_FLATPTR(regs->es, regs->bx);
     int status = send_disk_op(&dop);
     disk_ret(regs, status);
@@ -229,12 +209,11 @@ disk_1305(struct bregs *regs, struct drive_s *drive_g)
 
 // read disk drive parameters
 static void noinline
-disk_1308(struct bregs *regs, struct drive_s *drive_g)
+disk_1308(struct bregs *regs, struct drive_s *drive_gf)
 {
-    u16 ebda_seg = get_ebda_seg();
     // Get logical geometry from table
-    u16 nlc, nlh, nlspt;
-    fillLCHS(drive_g, &nlc, &nlh, &nlspt);
+    struct chs_s chs = getLCHS(drive_gf);
+    u16 nlc=chs.cylinder, nlh=chs.head, nls=chs.sector;
     nlc--;
     nlh--;
     u8 count;
@@ -242,11 +221,10 @@ disk_1308(struct bregs *regs, struct drive_s *drive_g)
         // Floppy
         count = GET_GLOBAL(FloppyCount);
 
-        if (CONFIG_CDROM_EMU
-            && drive_g == GLOBALFLAT2GLOBAL(GET_GLOBAL(cdemu_drive_gf)))
-            regs->bx = GET_EBDA2(ebda_seg, cdemu.media) * 2;
+        if (CONFIG_CDROM_EMU && drive_gf == GET_GLOBAL(cdemu_drive_gf))
+            regs->bx = GET_LOW(CDEmu.media) * 2;
         else
-            regs->bx = GET_GLOBAL(drive_g->floppy_type);
+            regs->bx = GET_GLOBALFLAT(drive_gf->floppy_type);
 
         // set es & di to point to 11 byte diskette param table in ROM
         regs->es = SEG_BIOS;
@@ -261,8 +239,8 @@ disk_1308(struct bregs *regs, struct drive_s *drive_g)
         return;
     }
 
-    if (CONFIG_CDROM_EMU && GET_EBDA2(ebda_seg, cdemu.active)) {
-        u8 emudrive = GET_EBDA2(ebda_seg, cdemu.emulated_extdrive);
+    if (CONFIG_CDROM_EMU && GET_LOW(CDEmu.active)) {
+        u8 emudrive = GET_LOW(CDEmu.emulated_extdrive);
         if (((emudrive ^ regs->dl) & 0x80) == 0)
             // Note extra drive due to emulation.
             count++;
@@ -273,7 +251,7 @@ disk_1308(struct bregs *regs, struct drive_s *drive_g)
 
     regs->al = 0;
     regs->ch = nlc & 0xff;
-    regs->cl = ((nlc >> 2) & 0xc0) | (nlspt & 0x3f);
+    regs->cl = ((nlc >> 2) & 0xc0) | (nls & 0x3f);
     regs->dh = nlh;
 
     disk_ret(regs, DISK_RET_SUCCESS);
@@ -282,33 +260,33 @@ disk_1308(struct bregs *regs, struct drive_s *drive_g)
 
 // initialize drive parameters
 static void
-disk_1309(struct bregs *regs, struct drive_s *drive_g)
+disk_1309(struct bregs *regs, struct drive_s *drive_gf)
 {
     DISK_STUB(regs);
 }
 
 // seek to specified cylinder
 static void
-disk_130c(struct bregs *regs, struct drive_s *drive_g)
+disk_130c(struct bregs *regs, struct drive_s *drive_gf)
 {
     DISK_STUB(regs);
 }
 
 // alternate disk reset
 static void
-disk_130d(struct bregs *regs, struct drive_s *drive_g)
+disk_130d(struct bregs *regs, struct drive_s *drive_gf)
 {
     DISK_STUB(regs);
 }
 
 // check drive ready
 static void
-disk_1310(struct bregs *regs, struct drive_s *drive_g)
+disk_1310(struct bregs *regs, struct drive_s *drive_gf)
 {
     // should look at 40:8E also???
 
     struct disk_op_s dop;
-    dop.drive_g = drive_g;
+    dop.drive_gf = drive_gf;
     dop.command = CMD_ISREADY;
     int status = send_disk_op(&dop);
     disk_ret(regs, status);
@@ -316,21 +294,21 @@ disk_1310(struct bregs *regs, struct drive_s *drive_g)
 
 // recalibrate
 static void
-disk_1311(struct bregs *regs, struct drive_s *drive_g)
+disk_1311(struct bregs *regs, struct drive_s *drive_gf)
 {
     DISK_STUB(regs);
 }
 
 // controller internal diagnostic
 static void
-disk_1314(struct bregs *regs, struct drive_s *drive_g)
+disk_1314(struct bregs *regs, struct drive_s *drive_gf)
 {
     DISK_STUB(regs);
 }
 
 // read disk drive size
 static void noinline
-disk_1315(struct bregs *regs, struct drive_s *drive_g)
+disk_1315(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret(regs, DISK_RET_SUCCESS);
     if (regs->dl < EXTSTART_HD || regs->dl >= EXTSTART_CD) {
@@ -341,18 +319,18 @@ disk_1315(struct bregs *regs, struct drive_s *drive_g)
     // Hard drive
 
     // Get logical geometry from table
-    u16 nlc, nlh, nlspt;
-    fillLCHS(drive_g, &nlc, &nlh, &nlspt);
+    struct chs_s chs = getLCHS(drive_gf);
+    u16 nlc=chs.cylinder, nlh=chs.head, nls=chs.sector;
 
     // Compute sector count seen by int13
-    u32 lba = (u32)(nlc - 1) * (u32)nlh * (u32)nlspt;
+    u32 lba = (u32)(nlc - 1) * (u32)nlh * (u32)nls;
     regs->cx = lba >> 16;
     regs->dx = lba & 0xffff;
     regs->ah = 3; // hard disk accessible
 }
 
 static void
-disk_1316(struct bregs *regs, struct drive_s *drive_g)
+disk_1316(struct bregs *regs, struct drive_s *drive_gf)
 {
     if (regs->dl >= EXTSTART_HD) {
         // Hard drive
@@ -364,7 +342,7 @@ disk_1316(struct bregs *regs, struct drive_s *drive_g)
 
 // IBM/MS installation check
 static void
-disk_1341(struct bregs *regs, struct drive_s *drive_g)
+disk_1341(struct bregs *regs, struct drive_s *drive_gf)
 {
     regs->bx = 0xaa55;  // install check
     regs->cx = 0x0007;  // ext disk access and edd, removable supported
@@ -374,79 +352,77 @@ disk_1341(struct bregs *regs, struct drive_s *drive_g)
 
 // IBM/MS extended read
 static void
-disk_1342(struct bregs *regs, struct drive_s *drive_g)
+disk_1342(struct bregs *regs, struct drive_s *drive_gf)
 {
-    extended_access(regs, drive_g, CMD_READ);
+    extended_access(regs, drive_gf, CMD_READ);
 }
 
 // IBM/MS extended write
 static void
-disk_1343(struct bregs *regs, struct drive_s *drive_g)
+disk_1343(struct bregs *regs, struct drive_s *drive_gf)
 {
-    extended_access(regs, drive_g, CMD_WRITE);
+    extended_access(regs, drive_gf, CMD_WRITE);
 }
 
 // IBM/MS verify
 static void
-disk_1344(struct bregs *regs, struct drive_s *drive_g)
+disk_1344(struct bregs *regs, struct drive_s *drive_gf)
 {
-    extended_access(regs, drive_g, CMD_VERIFY);
+    extended_access(regs, drive_gf, CMD_VERIFY);
 }
 
 // lock
 static void
-disk_134500(struct bregs *regs, struct drive_s *drive_g)
+disk_134500(struct bregs *regs, struct drive_s *drive_gf)
 {
-    u16 ebda_seg = get_ebda_seg();
     int cdid = regs->dl - EXTSTART_CD;
-    u8 locks = GET_EBDA2(ebda_seg, cdrom_locks[cdid]);
+    u8 locks = GET_LOW(CDRom_locks[cdid]);
     if (locks == 0xff) {
         regs->al = 1;
         disk_ret(regs, DISK_RET_ETOOMANYLOCKS);
         return;
     }
-    SET_EBDA2(ebda_seg, cdrom_locks[cdid], locks + 1);
+    SET_LOW(CDRom_locks[cdid], locks + 1);
     regs->al = 1;
     disk_ret(regs, DISK_RET_SUCCESS);
 }
 
 // unlock
 static void
-disk_134501(struct bregs *regs, struct drive_s *drive_g)
+disk_134501(struct bregs *regs, struct drive_s *drive_gf)
 {
-    u16 ebda_seg = get_ebda_seg();
     int cdid = regs->dl - EXTSTART_CD;
-    u8 locks = GET_EBDA2(ebda_seg, cdrom_locks[cdid]);
+    u8 locks = GET_LOW(CDRom_locks[cdid]);
     if (locks == 0x00) {
         regs->al = 0;
         disk_ret(regs, DISK_RET_ENOTLOCKED);
         return;
     }
     locks--;
-    SET_EBDA2(ebda_seg, cdrom_locks[cdid], locks);
+    SET_LOW(CDRom_locks[cdid], locks);
     regs->al = (locks ? 1 : 0);
     disk_ret(regs, DISK_RET_SUCCESS);
 }
 
 // status
 static void
-disk_134502(struct bregs *regs, struct drive_s *drive_g)
+disk_134502(struct bregs *regs, struct drive_s *drive_gf)
 {
     int cdid = regs->dl - EXTSTART_CD;
-    u8 locks = GET_EBDA(cdrom_locks[cdid]);
+    u8 locks = GET_LOW(CDRom_locks[cdid]);
     regs->al = (locks ? 1 : 0);
     disk_ret(regs, DISK_RET_SUCCESS);
 }
 
 static void
-disk_1345XX(struct bregs *regs, struct drive_s *drive_g)
+disk_1345XX(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret_unimplemented(regs, DISK_RET_EPARAM);
 }
 
 // IBM/MS lock/unlock drive
 static void
-disk_1345(struct bregs *regs, struct drive_s *drive_g)
+disk_1345(struct bregs *regs, struct drive_s *drive_gf)
 {
     if (regs->dl < EXTSTART_CD) {
         // Always success for HD
@@ -455,16 +431,16 @@ disk_1345(struct bregs *regs, struct drive_s *drive_g)
     }
 
     switch (regs->al) {
-    case 0x00: disk_134500(regs, drive_g); break;
-    case 0x01: disk_134501(regs, drive_g); break;
-    case 0x02: disk_134502(regs, drive_g); break;
-    default:   disk_1345XX(regs, drive_g); break;
+    case 0x00: disk_134500(regs, drive_gf); break;
+    case 0x01: disk_134501(regs, drive_gf); break;
+    case 0x02: disk_134502(regs, drive_gf); break;
+    default:   disk_1345XX(regs, drive_gf); break;
     }
 }
 
 // IBM/MS eject media
 static void noinline
-disk_1346(struct bregs *regs, struct drive_s *drive_g)
+disk_1346(struct bregs *regs, struct drive_s *drive_gf)
 {
     if (regs->dl < EXTSTART_CD) {
         // Volume Not Removable
@@ -473,7 +449,7 @@ disk_1346(struct bregs *regs, struct drive_s *drive_g)
     }
 
     int cdid = regs->dl - EXTSTART_CD;
-    u8 locks = GET_EBDA(cdrom_locks[cdid]);
+    u8 locks = GET_LOW(CDRom_locks[cdid]);
     if (locks != 0) {
         disk_ret(regs, DISK_RET_ELOCKED);
         return;
@@ -498,16 +474,18 @@ disk_1346(struct bregs *regs, struct drive_s *drive_g)
 
 // IBM/MS extended seek
 static void
-disk_1347(struct bregs *regs, struct drive_s *drive_g)
+disk_1347(struct bregs *regs, struct drive_s *drive_gf)
 {
-    extended_access(regs, drive_g, CMD_SEEK);
+    extended_access(regs, drive_gf, CMD_SEEK);
 }
 
 // IBM/MS get drive parameters
-static void
-disk_1348(struct bregs *regs, struct drive_s *drive_g)
+static void noinline
+disk_1348(struct bregs *regs, struct drive_s *drive_gf)
 {
-    u16 size = GET_INT13DPT(regs, size);
+    u16 seg = regs->ds;
+    struct int13dpt_s *param_far = (struct int13dpt_s*)(regs->si+0);
+    u16 size = GET_FARVAR(seg, param_far->size);
     u16 t13 = size == 74;
 
     // Buffer is too small
@@ -518,40 +496,40 @@ disk_1348(struct bregs *regs, struct drive_s *drive_g)
 
     // EDD 1.x
 
-    u8  type    = GET_GLOBAL(drive_g->type);
-    u16 npc     = GET_GLOBAL(drive_g->pchs.cylinders);
-    u16 nph     = GET_GLOBAL(drive_g->pchs.heads);
-    u16 npspt   = GET_GLOBAL(drive_g->pchs.spt);
-    u64 lba     = GET_GLOBAL(drive_g->sectors);
-    u16 blksize = GET_GLOBAL(drive_g->blksize);
+    u8  type    = GET_GLOBALFLAT(drive_gf->type);
+    u16 npc     = GET_GLOBALFLAT(drive_gf->pchs.cylinder);
+    u16 nph     = GET_GLOBALFLAT(drive_gf->pchs.head);
+    u16 nps     = GET_GLOBALFLAT(drive_gf->pchs.sector);
+    u64 lba     = GET_GLOBALFLAT(drive_gf->sectors);
+    u16 blksize = GET_GLOBALFLAT(drive_gf->blksize);
 
     dprintf(DEBUG_HDL_13, "disk_1348 size=%d t=%d chs=%d,%d,%d lba=%d bs=%d\n"
-            , size, type, npc, nph, npspt, (u32)lba, blksize);
+            , size, type, npc, nph, nps, (u32)lba, blksize);
 
-    SET_INT13DPT(regs, size, 26);
-    if (type == DTYPE_ATAPI) {
+    SET_FARVAR(seg, param_far->size, 26);
+    if (type == DTYPE_ATA_ATAPI) {
         // 0x74 = removable, media change, lockable, max values
-        SET_INT13DPT(regs, infos, 0x74);
-        SET_INT13DPT(regs, cylinders, 0xffffffff);
-        SET_INT13DPT(regs, heads, 0xffffffff);
-        SET_INT13DPT(regs, spt, 0xffffffff);
-        SET_INT13DPT(regs, sector_count, (u64)-1);
+        SET_FARVAR(seg, param_far->infos, 0x74);
+        SET_FARVAR(seg, param_far->cylinders, 0xffffffff);
+        SET_FARVAR(seg, param_far->heads, 0xffffffff);
+        SET_FARVAR(seg, param_far->spt, 0xffffffff);
+        SET_FARVAR(seg, param_far->sector_count, (u64)-1);
     } else {
-        if (lba > (u64)npspt*nph*0x3fff) {
-            SET_INT13DPT(regs, infos, 0x00); // geometry is invalid
-            SET_INT13DPT(regs, cylinders, 0x3fff);
+        if (lba > (u64)nps*nph*0x3fff) {
+            SET_FARVAR(seg, param_far->infos, 0x00); // geometry is invalid
+            SET_FARVAR(seg, param_far->cylinders, 0x3fff);
         } else {
-            SET_INT13DPT(regs, infos, 0x02); // geometry is valid
-            SET_INT13DPT(regs, cylinders, (u32)npc);
+            SET_FARVAR(seg, param_far->infos, 0x02); // geometry is valid
+            SET_FARVAR(seg, param_far->cylinders, (u32)npc);
         }
-        SET_INT13DPT(regs, heads, (u32)nph);
-        SET_INT13DPT(regs, spt, (u32)npspt);
-        SET_INT13DPT(regs, sector_count, lba);
+        SET_FARVAR(seg, param_far->heads, (u32)nph);
+        SET_FARVAR(seg, param_far->spt, (u32)nps);
+        SET_FARVAR(seg, param_far->sector_count, lba);
     }
-    SET_INT13DPT(regs, blksize, blksize);
+    SET_FARVAR(seg, param_far->blksize, blksize);
 
     if (size < 30 ||
-        (type != DTYPE_ATA && type != DTYPE_ATAPI &&
+        (type != DTYPE_ATA && type != DTYPE_ATA_ATAPI &&
          type != DTYPE_VIRTIO_BLK && type != DTYPE_VIRTIO_SCSI)) {
         disk_ret(regs, DISK_RET_SUCCESS);
         return;
@@ -563,19 +541,15 @@ disk_1348(struct bregs *regs, struct drive_s *drive_g)
     u16 iobase1 = 0;
     u64 device_path = 0;
     u8 channel = 0;
-    SET_INT13DPT(regs, size, 30);
-    if (type == DTYPE_ATA || type == DTYPE_ATAPI) {
-        u16 ebda_seg = get_ebda_seg();
-
-        SET_INT13DPT(regs, dpte_segment, ebda_seg);
-        SET_INT13DPT(regs, dpte_offset
-                     , offsetof(struct extended_bios_data_area_s, dpte));
+    SET_FARVAR(seg, param_far->size, 30);
+    if (type == DTYPE_ATA || type == DTYPE_ATA_ATAPI) {
+        SET_FARVAR(seg, param_far->dpte, SEGOFF(SEG_LOW, (u32)&DefaultDPTE));
 
         // Fill in dpte
-        struct atadrive_s *adrive_g = container_of(
-            drive_g, struct atadrive_s, drive);
-        struct ata_channel_s *chan_gf = GET_GLOBAL(adrive_g->chan_gf);
-        u8 slave = GET_GLOBAL(adrive_g->slave);
+        struct atadrive_s *adrive_gf = container_of(
+            drive_gf, struct atadrive_s, drive);
+        struct ata_channel_s *chan_gf = GET_GLOBALFLAT(adrive_gf->chan_gf);
+        u8 slave = GET_GLOBALFLAT(adrive_gf->slave);
         u16 iobase2 = GET_GLOBALFLAT(chan_gf->iobase2);
         u8 irq = GET_GLOBALFLAT(chan_gf->irq);
         iobase1 = GET_GLOBALFLAT(chan_gf->iobase1);
@@ -585,7 +559,7 @@ disk_1348(struct bregs *regs, struct drive_s *drive_g)
 
         u16 options = 0;
         if (type == DTYPE_ATA) {
-            u8 translation = GET_GLOBAL(drive_g->translation);
+            u8 translation = GET_GLOBALFLAT(drive_gf->translation);
             if (translation != TRANSLATION_NONE) {
                 options |= 1<<3; // CHS translation
                 if (translation == TRANSLATION_LBA)
@@ -602,26 +576,24 @@ disk_1348(struct bregs *regs, struct drive_s *drive_g)
         if (CONFIG_ATA_PIO32)
             options |= 1<<7;
 
-        SET_EBDA2(ebda_seg, dpte.iobase1, iobase1);
-        SET_EBDA2(ebda_seg, dpte.iobase2, iobase2 + ATA_CB_DC);
-        SET_EBDA2(ebda_seg, dpte.prefix, ((slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0)
-                                          | ATA_CB_DH_LBA));
-        SET_EBDA2(ebda_seg, dpte.unused, 0xcb);
-        SET_EBDA2(ebda_seg, dpte.irq, irq);
-        SET_EBDA2(ebda_seg, dpte.blkcount, 1);
-        SET_EBDA2(ebda_seg, dpte.dma, 0);
-        SET_EBDA2(ebda_seg, dpte.pio, 0);
-        SET_EBDA2(ebda_seg, dpte.options, options);
-        SET_EBDA2(ebda_seg, dpte.reserved, 0);
-        SET_EBDA2(ebda_seg, dpte.revision, 0x11);
+        SET_LOW(DefaultDPTE.iobase1, iobase1);
+        SET_LOW(DefaultDPTE.iobase2, iobase2 + ATA_CB_DC);
+        SET_LOW(DefaultDPTE.prefix, ((slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0)
+                                  | ATA_CB_DH_LBA));
+        SET_LOW(DefaultDPTE.unused, 0xcb);
+        SET_LOW(DefaultDPTE.irq, irq);
+        SET_LOW(DefaultDPTE.blkcount, 1);
+        SET_LOW(DefaultDPTE.dma, 0);
+        SET_LOW(DefaultDPTE.pio, 0);
+        SET_LOW(DefaultDPTE.options, options);
+        SET_LOW(DefaultDPTE.reserved, 0);
+        SET_LOW(DefaultDPTE.revision, 0x11);
 
-        u8 sum = checksum_far(
-            ebda_seg, (void*)offsetof(struct extended_bios_data_area_s, dpte), 15);
-        SET_EBDA2(ebda_seg, dpte.checksum, -sum);
+        u8 sum = checksum_far(SEG_LOW, &DefaultDPTE, 15);
+        SET_LOW(DefaultDPTE.checksum, -sum);
     } else {
-        SET_INT13DPT(regs, dpte_segment, 0);
-        SET_INT13DPT(regs, dpte_offset, 0);
-        bdf = GET_GLOBAL(drive_g->cntl_id);
+        SET_FARVAR(seg, param_far->dpte.segoff, 0xffffffff);
+        bdf = GET_GLOBALFLAT(drive_gf->cntl_id);
     }
 
     if (size < 66) {
@@ -630,60 +602,60 @@ disk_1348(struct bregs *regs, struct drive_s *drive_g)
     }
 
     // EDD 3.x
-    SET_INT13DPT(regs, key, 0xbedd);
-    SET_INT13DPT(regs, dpi_length, t13 ? 44 : 36);
-    SET_INT13DPT(regs, reserved1, 0);
-    SET_INT13DPT(regs, reserved2, 0);
+    SET_FARVAR(seg, param_far->key, 0xbedd);
+    SET_FARVAR(seg, param_far->dpi_length, t13 ? 44 : 36);
+    SET_FARVAR(seg, param_far->reserved1, 0);
+    SET_FARVAR(seg, param_far->reserved2, 0);
 
     if (bdf != -1) {
-        SET_INT13DPT(regs, host_bus[0], 'P');
-        SET_INT13DPT(regs, host_bus[1], 'C');
-        SET_INT13DPT(regs, host_bus[2], 'I');
-        SET_INT13DPT(regs, host_bus[3], ' ');
+        SET_FARVAR(seg, param_far->host_bus[0], 'P');
+        SET_FARVAR(seg, param_far->host_bus[1], 'C');
+        SET_FARVAR(seg, param_far->host_bus[2], 'I');
+        SET_FARVAR(seg, param_far->host_bus[3], ' ');
 
         u32 path = (pci_bdf_to_bus(bdf) | (pci_bdf_to_dev(bdf) << 8)
                     | (pci_bdf_to_fn(bdf) << 16));
         if (t13)
             path |= channel << 24;
 
-        SET_INT13DPT(regs, iface_path, path);
+        SET_FARVAR(seg, param_far->iface_path, path);
     } else {
         // ISA
-        SET_INT13DPT(regs, host_bus[0], 'I');
-        SET_INT13DPT(regs, host_bus[1], 'S');
-        SET_INT13DPT(regs, host_bus[2], 'A');
-        SET_INT13DPT(regs, host_bus[3], ' ');
+        SET_FARVAR(seg, param_far->host_bus[0], 'I');
+        SET_FARVAR(seg, param_far->host_bus[1], 'S');
+        SET_FARVAR(seg, param_far->host_bus[2], 'A');
+        SET_FARVAR(seg, param_far->host_bus[3], ' ');
 
-        SET_INT13DPT(regs, iface_path, iobase1);
+        SET_FARVAR(seg, param_far->iface_path, iobase1);
     }
 
     if (type != DTYPE_VIRTIO_BLK) {
-        SET_INT13DPT(regs, iface_type[0], 'A');
-        SET_INT13DPT(regs, iface_type[1], 'T');
-        SET_INT13DPT(regs, iface_type[2], 'A');
-        SET_INT13DPT(regs, iface_type[3], ' ');
+        SET_FARVAR(seg, param_far->iface_type[0], 'A');
+        SET_FARVAR(seg, param_far->iface_type[1], 'T');
+        SET_FARVAR(seg, param_far->iface_type[2], 'A');
+        SET_FARVAR(seg, param_far->iface_type[3], ' ');
     } else {
-        SET_INT13DPT(regs, iface_type[0], 'S');
-        SET_INT13DPT(regs, iface_type[1], 'C');
-        SET_INT13DPT(regs, iface_type[2], 'S');
-        SET_INT13DPT(regs, iface_type[3], 'I');
+        SET_FARVAR(seg, param_far->iface_type[0], 'S');
+        SET_FARVAR(seg, param_far->iface_type[1], 'C');
+        SET_FARVAR(seg, param_far->iface_type[2], 'S');
+        SET_FARVAR(seg, param_far->iface_type[3], 'I');
     }
-    SET_INT13DPT(regs, iface_type[4], ' ');
-    SET_INT13DPT(regs, iface_type[5], ' ');
-    SET_INT13DPT(regs, iface_type[6], ' ');
-    SET_INT13DPT(regs, iface_type[7], ' ');
+    SET_FARVAR(seg, param_far->iface_type[4], ' ');
+    SET_FARVAR(seg, param_far->iface_type[5], ' ');
+    SET_FARVAR(seg, param_far->iface_type[6], ' ');
+    SET_FARVAR(seg, param_far->iface_type[7], ' ');
 
     if (t13) {
-        SET_INT13DPT(regs, t13.device_path[0], device_path);
-        SET_INT13DPT(regs, t13.device_path[1], 0);
+        SET_FARVAR(seg, param_far->t13.device_path[0], device_path);
+        SET_FARVAR(seg, param_far->t13.device_path[1], 0);
 
-        SET_INT13DPT(regs, t13.checksum
-                     , -checksum_far(regs->ds, (void*)(regs->si+30), 43));
+        SET_FARVAR(seg, param_far->t13.checksum
+                   , -checksum_far(seg, (void*)param_far+30, 43));
     } else {
-        SET_INT13DPT(regs, phoenix.device_path, device_path);
+        SET_FARVAR(seg, param_far->phoenix.device_path, device_path);
 
-        SET_INT13DPT(regs, phoenix.checksum
-                     , -checksum_far(regs->ds, (void*)(regs->si+30), 35));
+        SET_FARVAR(seg, param_far->phoenix.checksum
+                   , -checksum_far(seg, (void*)param_far+30, 35));
     }
 
     disk_ret(regs, DISK_RET_SUCCESS);
@@ -691,7 +663,7 @@ disk_1348(struct bregs *regs, struct drive_s *drive_g)
 
 // IBM/MS extended media change
 static void
-disk_1349(struct bregs *regs, struct drive_s *drive_g)
+disk_1349(struct bregs *regs, struct drive_s *drive_gf)
 {
     if (regs->dl < EXTSTART_CD) {
         // Always success for HD
@@ -704,56 +676,56 @@ disk_1349(struct bregs *regs, struct drive_s *drive_g)
 }
 
 static void
-disk_134e01(struct bregs *regs, struct drive_s *drive_g)
+disk_134e01(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret(regs, DISK_RET_SUCCESS);
 }
 
 static void
-disk_134e03(struct bregs *regs, struct drive_s *drive_g)
+disk_134e03(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret(regs, DISK_RET_SUCCESS);
 }
 
 static void
-disk_134e04(struct bregs *regs, struct drive_s *drive_g)
+disk_134e04(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret(regs, DISK_RET_SUCCESS);
 }
 
 static void
-disk_134e06(struct bregs *regs, struct drive_s *drive_g)
+disk_134e06(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret(regs, DISK_RET_SUCCESS);
 }
 
 static void
-disk_134eXX(struct bregs *regs, struct drive_s *drive_g)
+disk_134eXX(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret(regs, DISK_RET_EPARAM);
 }
 
 // IBM/MS set hardware configuration
 static void
-disk_134e(struct bregs *regs, struct drive_s *drive_g)
+disk_134e(struct bregs *regs, struct drive_s *drive_gf)
 {
     switch (regs->al) {
-    case 0x01: disk_134e01(regs, drive_g); break;
-    case 0x03: disk_134e03(regs, drive_g); break;
-    case 0x04: disk_134e04(regs, drive_g); break;
-    case 0x06: disk_134e06(regs, drive_g); break;
-    default:   disk_134eXX(regs, drive_g); break;
+    case 0x01: disk_134e01(regs, drive_gf); break;
+    case 0x03: disk_134e03(regs, drive_gf); break;
+    case 0x04: disk_134e04(regs, drive_gf); break;
+    case 0x06: disk_134e06(regs, drive_gf); break;
+    default:   disk_134eXX(regs, drive_gf); break;
     }
 }
 
 static void
-disk_13XX(struct bregs *regs, struct drive_s *drive_g)
+disk_13XX(struct bregs *regs, struct drive_s *drive_gf)
 {
     disk_ret_unimplemented(regs, DISK_RET_EPARAM);
 }
 
 static void
-disk_13(struct bregs *regs, struct drive_s *drive_g)
+disk_13(struct bregs *regs, struct drive_s *drive_gf)
 {
     //debug_stub(regs);
 
@@ -761,37 +733,37 @@ disk_13(struct bregs *regs, struct drive_s *drive_g)
     SET_BDA(disk_interrupt_flag, 0);
 
     switch (regs->ah) {
-    case 0x00: disk_1300(regs, drive_g); break;
-    case 0x01: disk_1301(regs, drive_g); break;
-    case 0x02: disk_1302(regs, drive_g); break;
-    case 0x03: disk_1303(regs, drive_g); break;
-    case 0x04: disk_1304(regs, drive_g); break;
-    case 0x05: disk_1305(regs, drive_g); break;
-    case 0x08: disk_1308(regs, drive_g); break;
-    case 0x09: disk_1309(regs, drive_g); break;
-    case 0x0c: disk_130c(regs, drive_g); break;
-    case 0x0d: disk_130d(regs, drive_g); break;
-    case 0x10: disk_1310(regs, drive_g); break;
-    case 0x11: disk_1311(regs, drive_g); break;
-    case 0x14: disk_1314(regs, drive_g); break;
-    case 0x15: disk_1315(regs, drive_g); break;
-    case 0x16: disk_1316(regs, drive_g); break;
-    case 0x41: disk_1341(regs, drive_g); break;
-    case 0x42: disk_1342(regs, drive_g); break;
-    case 0x43: disk_1343(regs, drive_g); break;
-    case 0x44: disk_1344(regs, drive_g); break;
-    case 0x45: disk_1345(regs, drive_g); break;
-    case 0x46: disk_1346(regs, drive_g); break;
-    case 0x47: disk_1347(regs, drive_g); break;
-    case 0x48: disk_1348(regs, drive_g); break;
-    case 0x49: disk_1349(regs, drive_g); break;
-    case 0x4e: disk_134e(regs, drive_g); break;
-    default:   disk_13XX(regs, drive_g); break;
+    case 0x00: disk_1300(regs, drive_gf); break;
+    case 0x01: disk_1301(regs, drive_gf); break;
+    case 0x02: disk_1302(regs, drive_gf); break;
+    case 0x03: disk_1303(regs, drive_gf); break;
+    case 0x04: disk_1304(regs, drive_gf); break;
+    case 0x05: disk_1305(regs, drive_gf); break;
+    case 0x08: disk_1308(regs, drive_gf); break;
+    case 0x09: disk_1309(regs, drive_gf); break;
+    case 0x0c: disk_130c(regs, drive_gf); break;
+    case 0x0d: disk_130d(regs, drive_gf); break;
+    case 0x10: disk_1310(regs, drive_gf); break;
+    case 0x11: disk_1311(regs, drive_gf); break;
+    case 0x14: disk_1314(regs, drive_gf); break;
+    case 0x15: disk_1315(regs, drive_gf); break;
+    case 0x16: disk_1316(regs, drive_gf); break;
+    case 0x41: disk_1341(regs, drive_gf); break;
+    case 0x42: disk_1342(regs, drive_gf); break;
+    case 0x43: disk_1343(regs, drive_gf); break;
+    case 0x44: disk_1344(regs, drive_gf); break;
+    case 0x45: disk_1345(regs, drive_gf); break;
+    case 0x46: disk_1346(regs, drive_gf); break;
+    case 0x47: disk_1347(regs, drive_gf); break;
+    case 0x48: disk_1348(regs, drive_gf); break;
+    case 0x49: disk_1349(regs, drive_gf); break;
+    case 0x4e: disk_134e(regs, drive_gf); break;
+    default:   disk_13XX(regs, drive_gf); break;
     }
 }
 
 static void
-floppy_13(struct bregs *regs, struct drive_s *drive_g)
+floppy_13(struct bregs *regs, struct drive_s *drive_gf)
 {
     // Only limited commands are supported on floppies.
     switch (regs->ah) {
@@ -804,9 +776,9 @@ floppy_13(struct bregs *regs, struct drive_s *drive_g)
     case 0x08:
     case 0x15:
     case 0x16:
-        disk_13(regs, drive_g);
+        disk_13(regs, drive_gf);
         break;
-    default:   disk_13XX(regs, drive_g); break;
+    default:   disk_13XX(regs, drive_gf); break;
     }
 }
 
@@ -825,21 +797,21 @@ handle_legacy_disk(struct bregs *regs, u8 extdrive)
     }
 
     if (extdrive < EXTSTART_HD) {
-        struct drive_s *drive_g = getDrive(EXTTYPE_FLOPPY, extdrive);
-        if (!drive_g)
+        struct drive_s *drive_gf = getDrive(EXTTYPE_FLOPPY, extdrive);
+        if (!drive_gf)
             goto fail;
-        floppy_13(regs, drive_g);
+        floppy_13(regs, drive_gf);
         return;
     }
 
-    struct drive_s *drive_g;
+    struct drive_s *drive_gf;
     if (extdrive >= EXTSTART_CD)
-        drive_g = getDrive(EXTTYPE_CD, extdrive - EXTSTART_CD);
+        drive_gf = getDrive(EXTTYPE_CD, extdrive - EXTSTART_CD);
     else
-        drive_g = getDrive(EXTTYPE_HD, extdrive - EXTSTART_HD);
-    if (!drive_g)
+        drive_gf = getDrive(EXTTYPE_HD, extdrive - EXTSTART_HD);
+    if (!drive_gf)
         goto fail;
-    disk_13(regs, drive_g);
+    disk_13(regs, drive_gf);
     return;
 
 fail:
@@ -866,19 +838,17 @@ handle_13(struct bregs *regs)
             cdemu_134b(regs);
             return;
         }
-        u16 ebda_seg = get_ebda_seg();
-        if (GET_EBDA2(ebda_seg, cdemu.active)) {
-            u8 emudrive = GET_EBDA2(ebda_seg, cdemu.emulated_extdrive);
+        if (GET_LOW(CDEmu.active)) {
+            u8 emudrive = GET_LOW(CDEmu.emulated_extdrive);
             if (extdrive == emudrive) {
                 // Access to an emulated drive.
-                struct drive_s *cdemu_g;
-                cdemu_g = GLOBALFLAT2GLOBAL(GET_GLOBAL(cdemu_drive_gf));
+                struct drive_s *cdemu_gf = GET_GLOBAL(cdemu_drive_gf);
                 if (regs->ah > 0x16) {
                     // Only old-style commands supported.
-                    disk_13XX(regs, cdemu_g);
+                    disk_13XX(regs, cdemu_gf);
                     return;
                 }
-                disk_13(regs, cdemu_g);
+                disk_13(regs, cdemu_gf);
                 return;
             }
             if (extdrive < EXTSTART_CD && ((emudrive ^ extdrive) & 0x80) == 0)
@@ -895,7 +865,7 @@ handle_76(void)
 {
     debug_isr(DEBUG_ISR_76);
     SET_BDA(disk_interrupt_flag, 0xff);
-    eoi_pic2();
+    pic_eoi2();
 }
 
 // Old Fixed Disk Parameter Table (newer tables are in the ebda).
