@@ -4,33 +4,106 @@
 //
 // This file may be distributed under the terms of the GNU LGPLv3 license.
 
-#include "biosvar.h" // get_ebda_seg
-#include "util.h" // dprintf
+#include "biosvar.h" // GET_GLOBAL
 #include "bregs.h" // CR0_PE
+#include "hw/rtc.h" // rtc_use
+#include "list.h" // hlist_node
+#include "malloc.h" // free
+#include "output.h" // dprintf
+#include "stacks.h" // struct mutex_s
+#include "util.h" // useRTC
 
-// Thread info - stored at bottom of each thread stack - don't change
-// without also updating the inline assembler below.
-struct thread_info {
-    struct thread_info *next;
-    void *stackpos;
-    struct thread_info **pprev;
-};
-struct thread_info VAR32FLATVISIBLE MainThread = {
-    &MainThread, NULL, &MainThread.next
-};
+#define MAIN_STACK_MAX (1024*1024)
 
 
 /****************************************************************
- * Low level helpers
+ * Extra 16bit stack
  ****************************************************************/
 
-static inline void sgdt(struct descloc_s *desc) {
-    asm("sgdtl %0" : "=m"(*desc));
+// Space for a stack for 16bit code.
+u8 ExtraStack[BUILD_EXTRA_STACK_SIZE+1] VARLOW __aligned(8);
+u8 *StackPos VARLOW;
+
+// Test if currently on the extra stack
+static inline int
+on_extra_stack(void)
+{
+    return MODE16 && GET_SEG(SS) == SEG_LOW && getesp() > (u32)ExtraStack;
 }
-static inline void lgdt(struct descloc_s *desc) {
-    asm("lgdtl %0" : : "m"(*desc) : "memory");
+
+// Switch to the extra stack and call a function.
+u32
+stack_hop(u32 eax, u32 edx, void *func)
+{
+    if (on_extra_stack())
+        return ((u32 (*)(u32, u32))func)(eax, edx);
+    ASSERT16();
+    u16 stack_seg = SEG_LOW;
+    u32 bkup_ss, bkup_esp;
+    asm volatile(
+        // Backup current %ss/%esp values.
+        "movw %%ss, %w3\n"
+        "movl %%esp, %4\n"
+        // Copy stack seg to %ds/%ss and set %esp
+        "movw %w6, %%ds\n"
+        "movw %w6, %%ss\n"
+        "movl %5, %%esp\n"
+        "pushl %3\n"
+        "pushl %4\n"
+        // Call func
+        "calll *%2\n"
+        "popl %4\n"
+        "popl %3\n"
+        // Restore segments and stack
+        "movw %w3, %%ds\n"
+        "movw %w3, %%ss\n"
+        "movl %4, %%esp"
+        : "+a" (eax), "+d" (edx), "+c" (func), "=&r" (bkup_ss), "=&r" (bkup_esp)
+        : "m" (StackPos), "r" (stack_seg)
+        : "cc", "memory");
+    return eax;
 }
 
+// Switch back to original caller's stack and call a function.
+u32
+stack_hop_back(u32 eax, u32 edx, void *func)
+{
+    if (!on_extra_stack())
+        return ((u32 (*)(u32, u32))func)(eax, edx);
+    ASSERT16();
+    u16 bkup_ss;
+    u32 bkup_stack_pos, temp;
+    asm volatile(
+        // Backup stack_pos and current %ss/%esp
+        "movl %6, %4\n"
+        "movw %%ss, %w3\n"
+        "movl %%esp, %6\n"
+        // Restore original callers' %ss/%esp
+        "movl -4(%4), %5\n"
+        "movl %5, %%ss\n"
+        "movw %%ds:-8(%4), %%sp\n"
+        "movl %5, %%ds\n"
+        // Call func
+        "calll *%2\n"
+        // Restore %ss/%esp and stack_pos
+        "movw %w3, %%ds\n"
+        "movw %w3, %%ss\n"
+        "movl %6, %%esp\n"
+        "movl %4, %6"
+        : "+a" (eax), "+d" (edx), "+c" (func), "=&r" (bkup_ss)
+          , "=&r" (bkup_stack_pos), "=&r" (temp), "+m" (StackPos)
+        :
+        : "cc", "memory");
+    return eax;
+}
+
+
+/****************************************************************
+ * 16bit / 32bit calling
+ ****************************************************************/
+
+u16 StackSeg VARLOW;
+
 // Call a 32bit SeaBIOS function from a 16bit SeaBIOS function.
 u32 VISIBLE16
 call32(void *func, u32 eax, u32 errret)
@@ -51,6 +124,8 @@ call32(void *func, u32 eax, u32 errret)
     struct descloc_s gdt;
     sgdt(&gdt);
 
+    u16 oldstackseg = GET_LOW(StackSeg);
+    SET_LOW(StackSeg, GET_SEG(SS));
     u32 bkup_ss, bkup_esp;
     asm volatile(
         // Backup ss/esp / set esp to flat stack location
@@ -77,6 +152,8 @@ call32(void *func, u32 eax, u32 errret)
         : "r" (func)
         : "ecx", "edx", "cc", "memory");
 
+    SET_LOW(StackSeg, oldstackseg);
+
     // Restore gdt and fs/gs
     lgdt(&gdt);
     SET_SEG(FS, fs);
@@ -88,95 +165,76 @@ call32(void *func, u32 eax, u32 errret)
     return eax;
 }
 
-// 16bit trampoline for enabling irqs from 32bit mode.
-ASM16(
-    "  .global trampoline_checkirqs\n"
-    "trampoline_checkirqs:\n"
-    "  rep ; nop\n"
-    "  lretw"
-    );
-
-static void
-check_irqs(void)
+// Call a 16bit SeaBIOS function from a 32bit SeaBIOS function.
+static inline u32
+call16(u32 eax, u32 edx, void *func)
 {
-    if (MODESEGMENT) {
-        asm volatile(
-            "sti\n"
-            "nop\n"
-            "rep ; nop\n"
-            "cli\n"
-            "cld\n"
-            : : :"memory");
-        return;
-    }
-    extern void trampoline_checkirqs();
-    struct bregs br;
-    br.flags = F_IF;
-    br.code.seg = SEG_BIOS;
-    br.code.offset = (u32)&trampoline_checkirqs;
-    call16big(&br);
+    ASSERT32FLAT();
+    if (getesp() > MAIN_STACK_MAX)
+        panic("call16 with invalid stack\n");
+    extern u32 __call16(u32 eax, u32 edx, void *func);
+    return __call16(eax, edx, func - BUILD_BIOS_ADDR);
 }
 
-// 16bit trampoline for waiting for an irq from 32bit mode.
-ASM16(
-    "  .global trampoline_waitirq\n"
-    "trampoline_waitirq:\n"
-    "  sti\n"
-    "  hlt\n"
-    "  lretw"
-    );
-
-// Wait for next irq to occur.
-void
-wait_irq(void)
+static inline u32
+call16big(u32 eax, u32 edx, void *func)
 {
-    if (MODESEGMENT) {
-        asm volatile("sti ; hlt ; cli ; cld": : :"memory");
-        return;
-    }
-    if (CONFIG_THREADS && MainThread.next != &MainThread) {
-        // Threads still active - do a yield instead.
-        yield();
-        return;
-    }
-    extern void trampoline_waitirq();
-    struct bregs br;
-    br.flags = 0;
-    br.code.seg = SEG_BIOS;
-    br.code.offset = (u32)&trampoline_waitirq;
-    call16big(&br);
+    ASSERT32FLAT();
+    if (getesp() > MAIN_STACK_MAX)
+        panic("call16big with invalid stack\n");
+    extern u32 __call16big(u32 eax, u32 edx, void *func);
+    return __call16big(eax, edx, func - BUILD_BIOS_ADDR);
 }
 
 
 /****************************************************************
- * Stack in EBDA
+ * External 16bit interface calling
  ****************************************************************/
 
-// Switch to the extra stack in ebda and call a function.
-inline u32
-stack_hop(u32 eax, u32 edx, void *func)
+// Far call 16bit code with a specified register state.
+void VISIBLE16
+_farcall16(struct bregs *callregs, u16 callregseg)
 {
     ASSERT16();
-    u16 ebda_seg = get_ebda_seg(), bkup_ss;
-    u32 bkup_esp;
+    if (on_extra_stack()) {
+        stack_hop_back((u32)callregs, callregseg, _farcall16);
+        return;
+    }
     asm volatile(
-        // Backup current %ss/%esp values.
-        "movw %%ss, %w3\n"
-        "movl %%esp, %4\n"
-        // Copy ebda seg to %ds/%ss and set %esp
-        "movw %w6, %%ds\n"
-        "movw %w6, %%ss\n"
-        "movl %5, %%esp\n"
-        // Call func
-        "calll *%2\n"
-        // Restore segments and stack
-        "movw %w3, %%ds\n"
-        "movw %w3, %%ss\n"
-        "movl %4, %%esp"
-        : "+a" (eax), "+d" (edx), "+c" (func), "=&r" (bkup_ss), "=&r" (bkup_esp)
-        : "i" (EBDA_OFFSET_TOP_STACK), "r" (ebda_seg)
-        : "cc", "memory");
-    return eax;
+        "calll __farcall16\n"
+        : "+a" (callregs), "+m" (*callregs), "+d" (callregseg)
+        :
+        : "ebx", "ecx", "esi", "edi", "cc", "memory");
+}
+
+inline void
+farcall16(struct bregs *callregs)
+{
+    if (MODE16) {
+        _farcall16(callregs, GET_SEG(SS));
+        return;
+    }
+    extern void _cfunc16__farcall16(void);
+    call16((u32)callregs - StackSeg * 16, StackSeg, _cfunc16__farcall16);
+}
+
+inline void
+farcall16big(struct bregs *callregs)
+{
+    extern void _cfunc16__farcall16(void);
+    call16big((u32)callregs - StackSeg * 16, StackSeg, _cfunc16__farcall16);
+}
+
+// Invoke a 16bit software interrupt.
+inline void
+__call16_int(struct bregs *callregs, u16 offset)
+{
+    if (MODESEGMENT)
+        callregs->code.seg = GET_SEG(CS);
+    else
+        callregs->code.seg = SEG_BIOS;
+    callregs->code.offset = offset;
+    farcall16(callregs);
 }
 
 
@@ -184,15 +242,31 @@ stack_hop(u32 eax, u32 edx, void *func)
  * Threads
  ****************************************************************/
 
+// Thread info - stored at bottom of each thread stack - don't change
+// without also updating the inline assembler below.
+struct thread_info {
+    void *stackpos;
+    struct hlist_node node;
+};
+struct thread_info MainThread VARFSEG = {
+    NULL, { &MainThread.node, &MainThread.node.next }
+};
 #define THREADSTACKSIZE 4096
-int VAR16VISIBLE CanPreempt;
+
+// Check if any threads are running.
+static int
+have_threads(void)
+{
+    return (CONFIG_THREADS
+            && GET_FLATPTR(MainThread.node.next) != &MainThread.node);
+}
 
 // Return the 'struct thread_info' for the currently running thread.
 struct thread_info *
 getCurThread(void)
 {
     u32 esp = getesp();
-    if (esp <= BUILD_STACK_ADDR)
+    if (esp <= MAIN_STACK_MAX)
         return &MainThread;
     return (void*)ALIGN_DOWN(esp, THREADSTACKSIZE);
 }
@@ -201,15 +275,16 @@ getCurThread(void)
 static void
 switch_next(struct thread_info *cur)
 {
-    struct thread_info *next = cur->next;
+    struct thread_info *next = container_of(
+        cur->node.next, struct thread_info, node);
     if (cur == next)
         // Nothing to do.
         return;
     asm volatile(
         "  pushl $1f\n"                 // store return pc
         "  pushl %%ebp\n"               // backup %ebp
-        "  movl %%esp, 4(%%eax)\n"      // cur->stackpos = %esp
-        "  movl 4(%%ecx), %%esp\n"      // %esp = next->stackpos
+        "  movl %%esp, (%%eax)\n"       // cur->stackpos = %esp
+        "  movl (%%ecx), %%esp\n"       // %esp = next->stackpos
         "  popl %%ebp\n"                // restore %ebp
         "  retl\n"                      // restore pc
         "1:\n"
@@ -218,33 +293,14 @@ switch_next(struct thread_info *cur)
         : "ebx", "edx", "esi", "edi", "cc", "memory");
 }
 
-// Briefly permit irqs to occur.
-void
-yield(void)
-{
-    if (MODESEGMENT || !CONFIG_THREADS) {
-        // Just directly check irqs.
-        check_irqs();
-        return;
-    }
-    struct thread_info *cur = getCurThread();
-    if (cur == &MainThread)
-        // Permit irqs to fire
-        check_irqs();
-
-    // Switch to the next thread
-    switch_next(cur);
-}
-
 // Last thing called from a thread (called on "next" stack).
 static void
 __end_thread(struct thread_info *old)
 {
-    old->next->pprev = old->pprev;
-    *old->pprev = old->next;
+    hlist_del(&old->node);
     free(old);
     dprintf(DEBUG_thread, "\\%08x/ End thread\n", (u32)old);
-    if (MainThread.next == &MainThread)
+    if (!have_threads())
         dprintf(1, "All threads complete.\n");
 }
 
@@ -262,23 +318,20 @@ run_thread(void (*func)(void*), void *data)
 
     thread->stackpos = (void*)thread + THREADSTACKSIZE;
     struct thread_info *cur = getCurThread();
-    thread->next = cur;
-    thread->pprev = cur->pprev;
-    cur->pprev = &thread->next;
-    *thread->pprev = thread;
+    hlist_add_after(&thread->node, &cur->node);
 
     dprintf(DEBUG_thread, "/%08x\\ Start thread\n", (u32)thread);
     asm volatile(
         // Start thread
         "  pushl $1f\n"                 // store return pc
         "  pushl %%ebp\n"               // backup %ebp
-        "  movl %%esp, 4(%%edx)\n"      // cur->stackpos = %esp
-        "  movl 4(%%ebx), %%esp\n"      // %esp = thread->stackpos
+        "  movl %%esp, (%%edx)\n"       // cur->stackpos = %esp
+        "  movl (%%ebx), %%esp\n"       // %esp = thread->stackpos
         "  calll *%%ecx\n"              // Call func
 
         // End thread
-        "  movl (%%ebx), %%ecx\n"       // %ecx = thread->next
-        "  movl 4(%%ecx), %%esp\n"      // %esp = next->stackpos
+        "  movl 4(%%ebx), %%ecx\n"      // %ecx = thread->node.next
+        "  movl -4(%%ecx), %%esp\n"     // %esp = next->stackpos
         "  movl %%ebx, %%eax\n"
         "  calll %4\n"                  // call __end_thread(thread)
         "  popl %%ebp\n"                // restore %ebp
@@ -293,14 +346,77 @@ fail:
     func(data);
 }
 
+
+/****************************************************************
+ * Thread helpers
+ ****************************************************************/
+
+// Low-level irq enable.
+void VISIBLE16
+check_irqs(void)
+{
+    if (on_extra_stack()) {
+        stack_hop_back(0, 0, check_irqs);
+        return;
+    }
+    asm volatile("sti ; nop ; rep ; nop ; cli ; cld" : : :"memory");
+}
+
+// Briefly permit irqs to occur.
+void
+yield(void)
+{
+    if (MODESEGMENT) {
+        check_irqs();
+        return;
+    }
+    extern void _cfunc16_check_irqs(void);
+    if (!CONFIG_THREADS) {
+        call16big(0, 0, _cfunc16_check_irqs);
+        return;
+    }
+    struct thread_info *cur = getCurThread();
+    if (cur == &MainThread)
+        // Permit irqs to fire
+        call16big(0, 0, _cfunc16_check_irqs);
+
+    // Switch to the next thread
+    switch_next(cur);
+}
+
+void VISIBLE16
+wait_irq(void)
+{
+    if (on_extra_stack()) {
+        stack_hop_back(0, 0, wait_irq);
+        return;
+    }
+    asm volatile("sti ; hlt ; cli ; cld": : :"memory");
+}
+
+// Wait for next irq to occur.
+void
+yield_toirq(void)
+{
+    if (MODESEGMENT) {
+        wait_irq();
+        return;
+    }
+    if (have_threads()) {
+        // Threads still active - do a yield instead.
+        yield();
+        return;
+    }
+    extern void _cfunc16_wait_irq(void);
+    call16big(0, 0, _cfunc16_wait_irq);
+}
+
 // Wait for all threads (other than the main thread) to complete.
 void
 wait_threads(void)
 {
     ASSERT32FLAT();
-    if (! CONFIG_THREADS)
-        return;
-    while (MainThread.next != &MainThread)
+    while (have_threads())
         yield();
 }
 
@@ -329,29 +445,30 @@ mutex_unlock(struct mutex_s *mutex)
  * Thread preemption
  ****************************************************************/
 
+int CanPreempt VARFSEG;
 static u32 PreemptCount;
 
 // Turn on RTC irqs and arrange for them to check the 32bit threads.
 void
 start_preempt(void)
 {
-    if (! CONFIG_THREADS || ! CONFIG_THREAD_OPTIONROMS)
+    if (! CONFIG_THREAD_OPTIONROMS)
         return;
     CanPreempt = 1;
     PreemptCount = 0;
-    useRTC();
+    rtc_use();
 }
 
 // Turn off RTC irqs / stop checking for thread execution.
 void
 finish_preempt(void)
 {
-    if (! CONFIG_THREADS || ! CONFIG_THREAD_OPTIONROMS) {
+    if (! CONFIG_THREAD_OPTIONROMS) {
         yield();
         return;
     }
     CanPreempt = 0;
-    releaseRTC();
+    rtc_release();
     dprintf(9, "Done preempt - %d checks\n", PreemptCount);
     yield();
 }
@@ -360,8 +477,8 @@ finish_preempt(void)
 int
 wait_preempt(void)
 {
-    if (MODESEGMENT || !CONFIG_THREADS || !CONFIG_THREAD_OPTIONROMS
-        || !CanPreempt)
+    if (MODESEGMENT || !CONFIG_THREAD_OPTIONROMS || !CanPreempt
+        || getesp() < MAIN_STACK_MAX)
         return 0;
     while (CanPreempt)
         yield();
@@ -380,11 +497,7 @@ yield_preempt(void)
 void
 check_preempt(void)
 {
-    if (! CONFIG_THREADS || ! CONFIG_THREAD_OPTIONROMS
-        || !GET_GLOBAL(CanPreempt)
-        || GET_FLATPTR(MainThread.next) == &MainThread)
-        return;
-
     extern void _cfunc32flat_yield_preempt(void);
-    call32(_cfunc32flat_yield_preempt, 0, 0);
+    if (CONFIG_THREAD_OPTIONROMS && GET_GLOBAL(CanPreempt) && have_threads())
+        call32(_cfunc32flat_yield_preempt, 0, 0);
 }
