@@ -2,7 +2,6 @@
 #define __BOCHSVGA_H
 
 #include "types.h" // u8
-#include "ioport.h" // outw
 
 #define VBE_DISPI_BANK_ADDRESS           0xA0000
 #define VBE_DISPI_BANK_SIZE_KB           64
@@ -41,17 +40,6 @@
 
 #define VBE_DISPI_LFB_PHYSICAL_ADDRESS   0xE0000000
 
-static inline u16 dispi_read(u16 reg)
-{
-    outw(reg, VBE_DISPI_IOPORT_INDEX);
-    return inw(VBE_DISPI_IOPORT_DATA);
-}
-static inline void dispi_write(u16 reg, u16 val)
-{
-    outw(reg, VBE_DISPI_IOPORT_INDEX);
-    outw(val, VBE_DISPI_IOPORT_DATA);
-}
-
 struct vgamode_s *bochsvga_find_mode(int mode);
 void bochsvga_list_modes(u16 seg, u16 *dest, u16 *last);
 int bochsvga_get_window(struct vgamode_s *vmode_g, int window);
