@@ -5,10 +5,11 @@
 //
 // This file may be distributed under the terms of the GNU LGPLv3 license.
 
-#include "vgabios.h" // struct VideoParamTableEntry_s
 #include "biosvar.h" // GET_GLOBAL
-#include "util.h" // memcpy_far
+#include "output.h" // warn_internalerror
 #include "stdvga.h" // stdvga_find_mode
+#include "string.h" // memcpy_far
+#include "vgabios.h" // struct VideoParamTableEntry_s
 
 
 /****************************************************************
@@ -504,3 +505,10 @@ stdvga_set_mode(struct vgamode_s *vmode_g, int flags)
 
     return 0;
 }
+
+// Load the standard palette associated with 8bpp packed pixel vga modes.
+void
+stdvga_set_packed_palette(void)
+{
+    stdvga_dac_write(get_global_seg(), palette3, 0, sizeof(palette3) / 3);
+}
